#ifndef PHOTOMATIX32_H
#define PHOTOMATIX32_H

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#ifndef __unix
  #include  <io.h>
  #include  <direct.h>
  #ifndef _WIN32
    #define __stdcall
  #else
	#ifndef __cplusplus
		#define inline __inline
	#endif
  #endif
#else
  #include "WindowsEmulation.h"
#endif
#include <string.h>
#include <time.h>
#include <errno.h>
#include "MD5.h"
#ifndef __unix
#include <Windows.h>
#endif

#ifdef __unix
#include <pthread.h>
#else
#include "pthread.h"
#endif

#ifdef _WIN32
#define fseek64 _fseeki64
#define ftell64 _ftelli64
// VC 2008 has buggy fread/fwrite for sizes > 4GB, so added fread64/fwrite64 to read/write a chunk at a time
#else
#define fseek64 fseeko
#define ftell64 tfello
#define fread64 fread
#define fwrite64 fwrite
#endif


#if defined( __SSE__ )
#include <xmmintrin.h>
#endif

// Callback start
// Under windows, the callback is defined as __stdcall, as that seems to be required
// to call managed functions.  However when calling it from C/C++, the callback will
// also need to be __stdcall.
#ifdef _WIN32
typedef void (__stdcall *ProgressCallback)(float, int*);
#else
typedef void (*ProgressCallback)(float, int*);
#endif
// Callback end

#ifdef _DEBUG
#define ApplicationVersionMajor "4"
#define ApplicationVersionMinor "2"
#define ApplicationVersionRevision "PreBeta"
#else
#define ApplicationVersionMajor ".#VersionMajor#"
#define ApplicationVersionMinor ".#VersionMinor#"
#define ApplicationVersionRevision ".#VersionRevision#"
#endif

#ifndef __macos
#define __LITTLE_ENDIAN__
#endif
/*  .#cCloseComment#
.#DefTrace#
.#cOpenComment#  */
#define EXIFMAX 99999
#define MINRUN 4
#define TresPetit	-999999999.0
#define PlutotPetit	-99999999.0
#define TropGrand	100.0
#define TropPetit	-100.0
#define QuasiZero	1e-30
#define vPI 3.14159265358979323846
#define TresGrandF	999999999.0
#define fGrandeVal	99999999.0
#define GrandeVal	999999

#define XMP_PARAMS_LEN 18
#define XMP_LEN  60
#define TM_INFO_LEN 13
#define VAR_CLIP_ENHANCER (log(0.025)/log(0.3))

#define DEGHOST_DETECTION_HIGH 50
#define DEGHOST_DETECTION_NORMAL 70


// Value for red, green and blue that is used to make a grey colour to for areas of an image bitmap that are outside the bounds
// of the image.  The same value is used for red, green and blue.
#define NO_IMAGE_VALUE 196

#define IMAGE_INFO_LEN  13 

//100305 - EXIF0_LEN augmente de 2 pour embeddedXmp offset and len
//#define EXIF0_LEN  21
#define EXIF0_LEN  23

#define LENSINFO_LEN 64

#define MODEL_LEN 64
#define TIME_LEN 20

//#define EXIF_LEN  (EXIF0_LEN + ((MODEL_LEN+TIME_LEN)>>2)) 
#define EXIF_LEN  (EXIF0_LEN + ((MODEL_LEN+TIME_LEN+LENSINFO_LEN)>>2))

//#define RAW_INFO0_LEN  23
#define RAW_INFO0_LEN  (20 + ((3 + 3 + 1) * 2) + 1)

#define RAW_INFO_LEN  (RAW_INFO0_LEN + ((MODEL_LEN+TIME_LEN+LENSINFO_LEN)>>2) )

#define GPS0_LEN  ((6*3) + 2)
//6 pour stocker 3 rationals pour chacun des 3 tags suivants: lat_val, long_val et gps_timestamp
//2 pour stocker le rational de alt_val
#define GPS_LEN  (GPS0_LEN + 1 +3) 
// 1 pour stocker lat_ref, long_ref et alt_re 
// 3 pour stocker les 11 caracteres de gps_timestamp

#define IPTC_TAG_HEADER_LEN 5
#ifdef WIN32
typedef _int64 int64_t;
typedef unsigned _int64 u_int64_t;
#endif

#define LUM_PARAM_MAX 10

#ifdef _WIN32
#define rint(x) (floor((x)+.5))
#endif

// Use this instead for a format specifier to print size_t variables in printf like this:
// size_t x, y;
// printf("x=" SIZEFMT ", y=" SIZEFMT "\n", x, y);
#ifdef _WIN32
#define SIZETFMT "%Iu"
#else
#define SIZETFMT "%zu"
#endif

typedef struct {
    int nbProc;
	int readFromDisk;
	int methodType;
    const char *sourceFile;
    const char *tempFile0;
    const char *tempFile1;
    const char *tempFile2;
    const char *tempFile3;
    const char *tempFile4;
    const char *tempFile5;
} optim_param_struct;

typedef struct {
    int wImage;
	int hImage;
    int spp;
	int bpsIn;
	int bpsOut;
	int rowIn;
	int rowOut;
	int isBGR;
	int colorPlanes;
	int pluginPS;
} dim_param_struct;

// C versions of enums defined in C#.
// The numbers here must match the numbers given in the C# code.
// Enum literal names must be globally unique, so enum name is prefixed to each literal.
enum ENoiseReductionType
{
    ENoiseReductionType_None = -1,
    ENoiseReductionType_UnderExposed = 0,
    ENoiseReductionType_NormalAndUnderExposed = 1,
    ENoiseReductionType_All=2,
    ENoiseReductionType_MergedImage=10
};

enum EGhostingType
{
	EGhostingType_None = 0,
	EGhostingType_Automatic = 1,
	EGhostingType_SemiManual = 2
};

enum EAlignmentType
{
	EAlignmentType_None=-1,
	EAlignmentType_Translation=0,
	EAlignmentType_RotationPerspective=1,
	EAlignmentType_Rotation=2

};

enum EBlendingType
{
	EBlendingType_Auto=12,
	EBlendingType_Average=14,
	EBlendingType_TwoImages=13,
	EBlendingType_Adjust=10,
	EBlendingType_Intensive=11,
	EBlendingType_Realistic=15
};

enum EToneMappingMethod
{
	EToneMappingMethod_Unknown=0,
	EToneMappingMethod_DetailsEnhancer=1,
	EToneMappingMethod_ToneCompressor=2,
	EToneMappingMethod_ContrastOptimizer=3,
	EToneMappingMethod_ToneBalancer=4
};

enum EGhostingDetection
{
	EGhostingDetection_Normal=1,
	EGhostingDetection_High=2
};

#ifdef __cplusplus
extern "C" {
#endif
int __stdcall createDibDrawableBitmap(int w, int h, int srcRowStride, int destRowStride, int spp, int bps, float pixelAverage, const unsigned char *ratios, const float *exps2, const unsigned char *profileBuf, int profileLen, const char *profileName, const unsigned char *srcImage, unsigned char *destImage);
// profileName changed from char* to const char*
int  transformThruColorProfile(int w, int h, int spp, int stride, unsigned char destImage[], unsigned char srcImage[], unsigned char profileBuf[], int profileLen, const char* profileName);
// profileName changed from char* to const char*
int  __stdcall copyScaledBitmapFromBufferOptimized(int bmW, int bmH, int bfW, int bfH, int bmStride, int bfStride, int spp, int bps, int ox, int oy, int px, int py, float factor, float pixAverage, unsigned char bmImage[], unsigned char bfImage[], unsigned char ratios[], float exps2[], unsigned char profileBuf[], int profileLen, const char* profileName);
#ifdef __cplusplus
}
#endif
#endif
