#include "pmjp.h"

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef __unix
#include <jpeglib.h>
#else
#include "JpegSource/jpeglib.h"
#endif
#include "../Photomatix32.h"
#include "../PhotomatixMetaLib.h"

#include <typeinfo>

#define ICC_MARKER  (JPEG_APP0 + 2)    
#define ICC_OVERHEAD_LEN  14              
#define MAX_BYTES_IN_MARKER  65533   
#define MAX_DATA_BYTES_IN_MARKER  (MAX_BYTES_IN_MARKER - ICC_OVERHEAD_LEN)   

static int _debug=0, _debugThumb=0;


static boolean marker_is_icc (jpeg_saved_marker_ptr marker) {   
   return     marker->marker == ICC_MARKER && marker->data_length >= ICC_OVERHEAD_LEN &&     
   GETJOCTET(marker->data[0]) == 0x49 &&     GETJOCTET(marker->data[1]) == 0x43 &&    
   GETJOCTET(marker->data[2]) == 0x43 &&     GETJOCTET(marker->data[3]) == 0x5F &&    
   GETJOCTET(marker->data[4]) == 0x50 &&     GETJOCTET(marker->data[5]) == 0x52 &&   
   GETJOCTET(marker->data[6]) == 0x4F &&     GETJOCTET(marker->data[7]) == 0x46 &&   
   GETJOCTET(marker->data[8]) == 0x49 &&     GETJOCTET(marker->data[9]) == 0x4C &&    
   GETJOCTET(marker->data[10]) == 0x45 &&     GETJOCTET(marker->data[11]) == 0x0; 
}   

static void pm_error(j_common_ptr cinfo)
{
	throw cinfo->err->msg_code;
}

int __stdcall infojpg (const char *filename, int *info) {
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   FILE * infile = 0;		
//declarations pour recherche longueur profil 
   jpeg_saved_marker_ptr marker;  
   int num_markers=0, seq_no;   
   unsigned int total_length=0; 
   #define MAX_SEQ_NO  255             
   char marker_present[MAX_SEQ_NO+1];  
   unsigned int data_length[MAX_SEQ_NO+1]; 
   int retval;

if (_debug) (stderr,"infojpg - file: %s \n",filename);
   if ((infile = fopen(filename, "rb")) == NULL) return -1;

  cinfo.err = jpeg_std_error(&jerr);
  jerr.error_exit=pm_error;

  
  try
  {
	  jpeg_create_decompress(&cinfo);
		jpeg_save_markers(&cinfo,ICC_MARKER,0xFFFF);
	  jpeg_stdio_src(&cinfo, infile);
	  (void) jpeg_read_header(&cinfo, TRUE);
		   
	   jpeg_calc_output_dimensions(&cinfo);
		   
	  info[0] = cinfo.image_width;  info[1]= cinfo.image_height;  info[2]= cinfo.output_components;
	  info[3]=8;  //ajout 041108 
		info[4]=cinfo.X_density; info[5]=cinfo.Y_density;
	//recherche longueur profil pour allocation basÈ sur 1ere partie de read_icc_profile 
	   for (seq_no = 1; seq_no <= MAX_SEQ_NO; seq_no++)   marker_present[seq_no] = 0;    
	   for (marker=cinfo.marker_list; marker!= NULL; marker=marker->next) { 
   		if (marker_is_icc(marker)) {       
			 if (num_markers == 0)         num_markers = GETJOCTET(marker->data[13]);
			 else if (num_markers != GETJOCTET(marker->data[13]))  goto finRecherche;      
			 seq_no = GETJOCTET(marker->data[12]);      
	//		info[6]=seq_no;
			 if (seq_no <= 0 || seq_no > num_markers)    goto finRecherche;
			 if (marker_present[seq_no])         goto finRecherche;
			 marker_present[seq_no] = 1;       
	//		info[6]=marker->data_length;
			 data_length[seq_no] = marker->data_length - ICC_OVERHEAD_LEN;     
		}   
	   }    
	   if (num_markers == 0)  goto finRecherche;    
	   for (seq_no=1; seq_no<=num_markers; seq_no++) {     
			if (marker_present[seq_no] == 0) {total_length=0; goto finRecherche;} 
	//        data_offset[seq_no] = total_length;     
			total_length += data_length[seq_no];   
		}    
		if (total_length>0) info[6]=total_length;   
	finRecherche:
		retval = 8;
  }
  catch (int err)
  {
	  retval = -err;
  }
  jpeg_destroy_decompress(&cinfo);
  if (infile) fclose(infile);
// int i; for (i=0; i<7; i++) fprintf(stderr,"info[%d]=%d ",i,info[i]); fprintf(stderr,"\n");
  return retval;
}

int __stdcall jpg_profile (const char *filename, unsigned char *icc_data) {
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   jpeg_saved_marker_ptr marker;  
   int num_markers=0, seq_no;   unsigned int total_length=0; 
   #define MAX_SEQ_NO  255             
   char marker_present[MAX_SEQ_NO+1];  
   unsigned int data_length[MAX_SEQ_NO+1], data_offset[MAX_SEQ_NO+1], length; 
   unsigned char *src_ptr, *dst_ptr;                         
   int retval;
  
	FILE * infile = 0;		
	if ((infile = fopen(filename, "rb")) == NULL) return -1;
	
	cinfo.err = jpeg_std_error(&jerr);
	jerr.error_exit=pm_error;

    try
    {
		jpeg_create_decompress(&cinfo);
		jpeg_save_markers(&cinfo,ICC_MARKER,0xFFFF);
		jpeg_stdio_src(&cinfo, infile);
		(void) jpeg_read_header(&cinfo, TRUE);

		for (seq_no = 1; seq_no <= MAX_SEQ_NO; seq_no++)   marker_present[seq_no] = 0;    
		for (marker=cinfo.marker_list; marker!= NULL; marker=marker->next) { 
			if (marker_is_icc(marker)) {       
				if (num_markers == 0)         num_markers = GETJOCTET(marker->data[13]);
				else if (num_markers != GETJOCTET(marker->data[13]))  goto finRecherche;      
				seq_no = GETJOCTET(marker->data[12]);      
				if (seq_no <= 0 || seq_no > num_markers)    goto finRecherche;
				if (marker_present[seq_no])         goto finRecherche;
				marker_present[seq_no] = 1;       
				data_length[seq_no] = marker->data_length - ICC_OVERHEAD_LEN;     
			}   
	   }    
	   if (num_markers == 0)  goto finRecherche;    
	   for (seq_no=1; seq_no<=num_markers; seq_no++) {     
			if (marker_present[seq_no] == 0) {total_length=0; goto finRecherche;} 
			data_offset[seq_no] = total_length;     
			total_length += data_length[seq_no];   
	//fprintf(stderr,"seq_no=%d  total_len=%d\n",seq_no,total_length);
		}    
		if (total_length <= 0)   goto finRecherche;
		for (marker = cinfo.marker_list; marker != NULL; marker = marker->next) {    
			if (marker_is_icc(marker)) {       
				seq_no = GETJOCTET(marker->data[12]);
				dst_ptr = icc_data + data_offset[seq_no];       
				src_ptr = marker->data + ICC_OVERHEAD_LEN;       
				length = data_length[seq_no];       
				while (length--) {         
					*dst_ptr++ = *src_ptr++;       
				}     
			}   
		}    
	finRecherche:
		retval = total_length;
  }
  catch (int err)
  {
	  retval = -err;
  }

  jpeg_destroy_decompress(&cinfo);
  if (infile) fclose(infile);
  return retval;
}


int __stdcall PMwritejpg (const char *filename, int quality, int *info, int row_stride, unsigned char *icc_data, int *exifBuf, int *gpsBuf, int iptc_len, unsigned char *iptc_data, int tmMethod, float *tmParam, unsigned char *pImage) {
	int image_width=info[0], image_height=info[1], spp=info[2], icc_data_len=info[6];
	int xres=info[4], yres=info[5], is16to8=info[7], orientation=info[10];
	int retour=0, exifCount=0, exifDataLen=0, gpsCount=0, gpsDataLen=0, modelLen=0, makeLen=0, datetimeLen=0, gpsDatestampLen=0, ifdtotal=0, lensLen=0; //100308
	int nifds=0, exifTotalLen=0, gpsTotalLen=0, exifIfdLen=0, gpsIfdLen=0, xmpLen=0;  //100307
	unsigned char *exif_data_ptr, *iptc_data_ptr;  char *xmpData, *xmp_data_ptr, *lens=NULL, *pch;  //100307
	char pMakeInfo[MODEL_LEN];  //120925
	int retval;

	JSAMPLE *pImg=pImage;
	struct jpeg_compress_struct cinfo;	struct jpeg_error_mgr jerr;
	FILE *outfile;		  
	unsigned char *scanlineBuf=NULL;  int posit16=0, i=0; unsigned short *p16;
	
	if ((outfile = fopen(filename, "wb")) == NULL) {fprintf(stderr, "can't open %s\n", filename); return -1; } 
	cinfo.err = jpeg_std_error(&jerr); 
	jerr.error_exit=pm_error;
	try
	{
		jpeg_create_compress(&cinfo);
		jpeg_stdio_dest(&cinfo, outfile);

		cinfo.image_width = image_width; cinfo.image_height = image_height;  cinfo.input_components = spp;	
		if (spp>1) cinfo.in_color_space = JCS_RGB; 	
		else cinfo.in_color_space = JCS_GRAYSCALE; 
		jpeg_set_defaults(&cinfo);
		cinfo.X_density=xres; cinfo.Y_density=yres;  cinfo.density_unit=1;
		jpeg_set_quality(&cinfo, quality, TRUE);
		//  jpeg_set_marker_processor(&cinfo, quality, TRUE);
		jpeg_start_compress(&cinfo, TRUE);

		/*
		SOI  - start of image         (FFD8)
		APP0						  (FFE0)
		APP1 - EXIF segment           (FFE1)  metadata like: description, date/time taken, author, camera details
		APP2 - EXIF extension         (FFE2)  photo quality details, GPS details, thumbnail etc 
		COM  - comments segment       (FFFE) 
		APP13 - IPTC segment          (FFED) 
		SOF  - Start of Frame segment (FFC0)
		*/

		//ecriture EXIF et GPS data dans APP1
		if (exifBuf == NULL && gpsBuf == NULL)
		{
			exifCount = gpsCount = 0;
		}
		else
		{
			retour=get_exif_count(exifBuf, &exifCount, &exifDataLen, gpsBuf, &gpsCount, &gpsDataLen, &modelLen, &makeLen, &datetimeLen, &gpsDatestampLen, &lensLen, pMakeInfo);  //100308
		}
		if (_debug)	fprintf(stderr,"exifCount:%d    exifDataLen:%d   lensInfo: %d \n",exifCount, exifDataLen,lensLen);

		exifIfdLen= 2 + 12*exifCount + 4; gpsIfdLen= 2 + 12*gpsCount+ 4; 
		if (modelLen>0) nifds++;  if (makeLen>0) nifds++; if (exifCount>0) nifds++; if (gpsCount>0) nifds++;	
		nifds++; //090810 pour orientation qui est systematiquement ajoutée (forcée à 1 si non renseignée)
		if (nifds>1) {  //090822 if (nifds>0) {
			exifTotalLen=(exifCount>0)?exifIfdLen+exifDataLen:0; gpsTotalLen=(gpsCount>0)?gpsIfdLen+gpsDataLen:0;
			ifdtotal = 8 + (2+ 12*nifds + 4) + modelLen + makeLen + exifTotalLen + gpsTotalLen;  

			jpeg_write_m_header(&cinfo, 0xE1, (unsigned int) (4 + 2 + ifdtotal));   
			jpeg_write_m_byte(&cinfo, 0x45); jpeg_write_m_byte(&cinfo, 0x78);  jpeg_write_m_byte(&cinfo, 0x69); jpeg_write_m_byte(&cinfo, 0x66); //Exif    
			jpeg_write_m_byte(&cinfo, 0x0);  jpeg_write_m_byte(&cinfo, 0x0);      
			//		jpeg_write_m_byte(&cinfo, 0x4D);  jpeg_write_m_byte(&cinfo, 0x4D); //MM     
			//		jpeg_write_m_byte(&cinfo, 0x0);  jpeg_write_m_byte(&cinfo, 0x2A);  //42 en endianess M      
			//		jpeg_write_m_byte(&cinfo, 0x0);  jpeg_write_m_byte(&cinfo, 0x0);   jpeg_write_m_byte(&cinfo, 0x0);  jpeg_write_m_byte(&cinfo, 0x08);      


			unsigned char *exif_data_base = (unsigned char*)calloc(ifdtotal+1, sizeof(unsigned char));
			exif_data_ptr=exif_data_base;
			if (exif_data_ptr!=NULL) {
				retour=write_exif_buf(exifCount, exifDataLen, gpsCount, gpsDataLen, modelLen, makeLen, datetimeLen, gpsDatestampLen, /*090910*/orientation, exifBuf, gpsBuf, pMakeInfo,  exif_data_ptr);
				if (_debug)			fprintf(stderr, "retour write_exif_buf: %d  idftotal: %d\n", retour, ifdtotal);
				if (retour==ifdtotal) {
					while (ifdtotal--) {
						//if (_debug) if ((ifdtotal%16)==0) fprintf(stderr, "write_m_byte: %d\n", ifdtotal);       
						jpeg_write_m_byte(&cinfo, *exif_data_ptr);       
						exif_data_ptr++;
					}
				}
				free(exif_data_base);
			}
		}

		//ecriture IPTC metadata dans APP13 (Photoshop IRB)
		if (iptc_len>0) {
			int addAlignmentByte=iptc_len%2;  //090804
			if (_debug) fprintf(stderr,"iptc_len: %d   addAlignmentByte: %d\n",iptc_len,addAlignmentByte);

			//14 pour Photoshop 3.0\0, 4 pour 8BIM, 2 pour iptc marker (0404), 2 pour 00 et 4 pour longeur (iptc_len)
			jpeg_write_m_header(&cinfo, 0xED, (unsigned int) (14 + 4 + 2 +2 +4 +iptc_len+addAlignmentByte));   //090804

			jpeg_write_m_byte(&cinfo, 0x50); jpeg_write_m_byte(&cinfo, 0x68);  jpeg_write_m_byte(&cinfo, 0x6F); jpeg_write_m_byte(&cinfo, 0x74);    
			jpeg_write_m_byte(&cinfo, 0x6F);  jpeg_write_m_byte(&cinfo, 0x73);  jpeg_write_m_byte(&cinfo, 0x68);  jpeg_write_m_byte(&cinfo, 0x6F);      
			jpeg_write_m_byte(&cinfo, 0x70);  jpeg_write_m_byte(&cinfo, 0x20);  jpeg_write_m_byte(&cinfo, 0x33);  jpeg_write_m_byte(&cinfo, 0x2E);      
			jpeg_write_m_byte(&cinfo, 0x30);  jpeg_write_m_byte(&cinfo, 0x0);  
			jpeg_write_m_byte(&cinfo, 0x38);  jpeg_write_m_byte(&cinfo, 0x42);  jpeg_write_m_byte(&cinfo, 0x49);  jpeg_write_m_byte(&cinfo, 0x4D);      
			jpeg_write_m_byte(&cinfo, 0x04);  jpeg_write_m_byte(&cinfo, 0x04);  
			jpeg_write_m_byte(&cinfo, 0x0);  jpeg_write_m_byte(&cinfo, 0x0);      

			if (_debug) fprintf(stderr, "retour decomp hex iptc_len: %x %x %x %x \n", iptc_len>>24,iptc_len>>16,iptc_len>>8,iptc_len);

			//080919 tentative fix bug incorrectly written IPTC data on PPC
			jpeg_write_m_byte(&cinfo, iptc_len>>24); jpeg_write_m_byte(&cinfo, iptc_len>>16); jpeg_write_m_byte(&cinfo, iptc_len>>8); jpeg_write_m_byte(&cinfo, iptc_len); 
			/*	#if defined(__LITTLE_ENDIAN__)
			jpeg_write_m_byte(&cinfo, iptc_len>>24); jpeg_write_m_byte(&cinfo, iptc_len>>16); jpeg_write_m_byte(&cinfo, iptc_len>>8); jpeg_write_m_byte(&cinfo, iptc_len); 
			#else
			jpeg_write_m_byte(&cinfo, iptc_len); jpeg_write_m_byte(&cinfo, iptc_len>>8); jpeg_write_m_byte(&cinfo, iptc_len>>16); jpeg_write_m_byte(&cinfo, iptc_len>>24); 
			#endif
			*/
			//080919 - fin tentative fix bug incorrectly written IPTC data on PPC
			iptc_data_ptr=iptc_data;
			while (iptc_len--) {
				jpeg_write_m_byte(&cinfo, *iptc_data_ptr);     
				iptc_data_ptr++;
			}
			if (addAlignmentByte>0)  jpeg_write_m_byte(&cinfo, 0x0);   //090804 - rajout pour alignement afin d'eviter erreur a ouverture dans Photoshop quand longueur data iptc est impaire
		}


	if ((tmMethod>0)||(lensLen>0)) {  //ajout XMP data dans APP1

//100619 - fix	start			
		lens=(char*)calloc((lensLen>0)?lensLen:1, 1);  	
		pch=(char*)exifBuf;  
		if (lensLen>0)	memcpy(lens, pch+(EXIF0_LEN*sizeof(int))+MODEL_LEN+TIME_LEN, lensLen);
//100619 - fix	end

		xmpLen=embed_xmp(0, tmMethod, tmParam, lensLen, lens, NULL); if (_debug) fprintf(stderr, "tmMethod: %d    xmpLen: %d\n", tmMethod, xmpLen); 
		
		if (xmpLen>0) {
			xmpData=(char*)calloc(xmpLen+1, sizeof(char));  
			if (xmpData!=NULL) {
				jpeg_write_m_header(&cinfo, 0xE1, (unsigned int) (29 + xmpLen));   //29 pour http://ns.adobe.com/xap/1.0/\0
			
				jpeg_write_m_byte(&cinfo, 0x68); jpeg_write_m_byte(&cinfo, 0x74); jpeg_write_m_byte(&cinfo, 0x74); jpeg_write_m_byte(&cinfo, 0x70); jpeg_write_m_byte(&cinfo, 0x3A);
				jpeg_write_m_byte(&cinfo, 0x2F); jpeg_write_m_byte(&cinfo, 0x2F); jpeg_write_m_byte(&cinfo, 0x6E); jpeg_write_m_byte(&cinfo, 0x73); jpeg_write_m_byte(&cinfo, 0x2E);
				jpeg_write_m_byte(&cinfo, 0x61); jpeg_write_m_byte(&cinfo, 0x64); jpeg_write_m_byte(&cinfo, 0x6F); jpeg_write_m_byte(&cinfo, 0x62); jpeg_write_m_byte(&cinfo, 0x65);
				jpeg_write_m_byte(&cinfo, 0x2E); jpeg_write_m_byte(&cinfo, 0x63); jpeg_write_m_byte(&cinfo, 0x6F); jpeg_write_m_byte(&cinfo, 0x6D); jpeg_write_m_byte(&cinfo, 0x2F);
				jpeg_write_m_byte(&cinfo, 0x78); jpeg_write_m_byte(&cinfo, 0x61); jpeg_write_m_byte(&cinfo, 0x70); jpeg_write_m_byte(&cinfo, 0x2F); jpeg_write_m_byte(&cinfo, 0x31);
				jpeg_write_m_byte(&cinfo, 0x2E); jpeg_write_m_byte(&cinfo, 0x30); jpeg_write_m_byte(&cinfo, 0x2F); 
				jpeg_write_m_byte(&cinfo, 0x0); 

//100619 - fix	start			
/*
				pch=(char*)exifBuf;  lens=calloc(lensLen, 1); 
				if (lens==NULL) lensLen=0;
				else memcpy(lens, pch+(EXIF0_LEN*sizeof(int))+MODEL_LEN+TIME_LEN, lensLen);
				free(lens);
*/
//100619 - fix	end			

				retour=embed_xmp(1, tmMethod, tmParam, lensLen, lens, xmpData);  		if (_debug) fprintf(stderr, "retour embed_xmp: %d\n", retour); 

				xmp_data_ptr=xmpData;
				while (xmpLen--) {
					jpeg_write_m_byte(&cinfo, *xmp_data_ptr);     
					xmp_data_ptr++;
				}			
			}
			free(xmpData);
		}
		free(lens); //100619 fix
	}


		//ecriture ICC profile dans APP2
		if (icc_data_len>0) {
			unsigned char *icc_data_ptr=icc_data;
			unsigned int num_markers;    // total number of markers we'll write    
			int cur_marker = 1;           // per spec, counting starts at 1    
			unsigned int length;            // number of bytes to write in this marker     
			num_markers = icc_data_len / MAX_DATA_BYTES_IN_MARKER;   
			if (num_markers * MAX_DATA_BYTES_IN_MARKER != icc_data_len)     num_markers++;    
			while (icc_data_len > 0) {      
				length = icc_data_len;     
				if (length > MAX_DATA_BYTES_IN_MARKER)       length = MAX_DATA_BYTES_IN_MARKER;    
				icc_data_len -= length;      
				jpeg_write_m_header(&cinfo, ICC_MARKER, (unsigned int) (length + ICC_OVERHEAD_LEN));   
				jpeg_write_m_byte(&cinfo, 0x49);     
				jpeg_write_m_byte(&cinfo, 0x43);    jpeg_write_m_byte(&cinfo, 0x43);     
				jpeg_write_m_byte(&cinfo, 0x5F);     jpeg_write_m_byte(&cinfo, 0x50);     
				jpeg_write_m_byte(&cinfo, 0x52);     jpeg_write_m_byte(&cinfo, 0x4F);     
				jpeg_write_m_byte(&cinfo, 0x46);     jpeg_write_m_byte(&cinfo, 0x49);     
				jpeg_write_m_byte(&cinfo, 0x4C);     jpeg_write_m_byte(&cinfo, 0x45);    
				jpeg_write_m_byte(&cinfo, 0x0);      
				jpeg_write_m_byte(&cinfo, cur_marker);     jpeg_write_m_byte(&cinfo, (int) num_markers);      
				while (length--) {       
					jpeg_write_m_byte(&cinfo, *icc_data_ptr);       
					icc_data_ptr++;     
				}
				cur_marker++;   
			} 
		}

		if (is16to8==0) {
			while (cinfo.next_scanline < cinfo.image_height) {
				//if (_debug)	fprintf(stderr,"0: image_height: %d    row_stride: %d  \n",cinfo.image_height, row_stride);
				(void) jpeg_write_scanlines(&cinfo, &pImg, 1);
				pImg += row_stride;  
			}
		}
		else if (is16to8==1) {
			//if (_debug)	fprintf(stderr,"1: image_height: %d    row_stride: %d  \n",cinfo.image_height, row_stride);
			scanlineBuf=(unsigned char*)calloc(row_stride, sizeof(unsigned char));
			p16=(unsigned short*)pImage; posit16=0;
			while (cinfo.next_scanline < cinfo.image_height) {
				//			if (_debug) if ((cinfo.next_scanline%500)==0) fprintf(stderr,"cinfo.next_scanline=%d \n",cinfo.next_scanline);
				for (i=0; i<row_stride; i++) {
					scanlineBuf[i]=(p16[posit16])>>8;
					posit16++;
				}
				(void) jpeg_write_scanlines(&cinfo, &scanlineBuf, 1);
			}
			free(scanlineBuf);
		}
		jpeg_finish_compress(&cinfo);
		fclose(outfile);
		jpeg_destroy_compress(&cinfo);

	}
	catch (int err)
	{
		retval = -err;
	}

	return cinfo.next_scanline;
}

int  __stdcall PMreadjpgflat (const char *filename, unsigned char *bufImage) {

  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  FILE * infile = 0;		
  unsigned char *buf8;
  int ligne=0, spp, iMax, rowstride,retval=0;  

  if ((infile = fopen(filename, "rb")) == NULL) {
    fprintf(stderr, "can't open %s\n", filename);
    return -1;
  }
//  fprintf(stderr, "pmjp: %s\n", filename);

  cinfo.err = jpeg_std_error(&jerr);
  jerr.error_exit=pm_error;

  try
  {
	  jpeg_create_decompress(&cinfo);
	  jpeg_stdio_src(&cinfo, infile);
	  
	  retval= jpeg_read_header(&cinfo, TRUE); 
	//  fprintf(stderr, "pmjp: read_header %d\n",retval);
	  retval= jpeg_start_decompress(&cinfo);
	//  fprintf(stderr, "pmjp: start decomp %d\n",retval);

	  iMax = cinfo.output_width;
	  spp=cinfo.output_components;
	  rowstride=iMax*spp;
	  buf8=(unsigned char*)calloc(rowstride,sizeof(unsigned char));  

	  //fprintf(stderr, "pmjp: rowstride: %d\n", rowstride);

	   while (cinfo.output_scanline < cinfo.output_height) {
			(void) jpeg_read_scanlines(&cinfo, &buf8, 1);
	//	  p8=buf8;
		  memcpy (bufImage+ligne, buf8, rowstride); //for (i=0; i<rowstride; i++) bufImage[i]=p8[i];
		  ligne+=rowstride;
	   }
	//  fprintf(stderr, "pmjp: ligne: %d\n", ligne);

	   free(buf8);  
	  retval= jpeg_finish_decompress(&cinfo);
	  //fprintf(stderr, "pmjp: finish %d\n",retval);
	  retval=cinfo.output_scanline;
  }
  catch (int err)
  {
	  retval = -err;
  }
  jpeg_destroy_decompress(&cinfo);
  if (infile) fclose(infile);
  return retval;
}



/////////////////////// readjpg4dib /////////////////////////////
int  __stdcall PMreadjpg4dib (const char *filename, int rowOut, unsigned char *bufImage) {

  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  FILE * infile = 0;		
  unsigned char *p8, *buf8;
  int i,spp, iMax, retval=0,ifin;  int c1=1, c2=2;

  if ((infile = fopen(filename, "rb")) == NULL) {
    fprintf(stderr, "can't open %s\n", filename);
    return -1;
  }

  cinfo.err = jpeg_std_error(&jerr);
  jerr.error_exit=pm_error;

  try
  {
	  jpeg_create_decompress(&cinfo);
	  jpeg_stdio_src(&cinfo, infile);

	  (void) jpeg_read_header(&cinfo, TRUE);   
	  (void) jpeg_start_decompress(&cinfo);

	  iMax = cinfo.output_width;
	  spp=cinfo.output_components;  if (spp==1) c1=c2=0;
	  ifin=iMax*spp;
	  buf8=(unsigned char*)calloc(ifin,sizeof(unsigned char));  

	  if (rowOut>ifin) {
		 while (cinfo.output_scanline < cinfo.output_height) {
			(void) jpeg_read_scanlines(&cinfo, &buf8, 1);
		  p8=buf8;
			for (i=0; i<iMax;i++) {	
			bufImage[0]=p8[c2];
			bufImage[c1]=p8[c1];
			bufImage[c2]=p8[0];
			bufImage+=spp; p8+=spp;
		  }
			for (i=ifin; i<rowOut; i++) bufImage++;
		 }
	  }
	  else {
		 while (cinfo.output_scanline < cinfo.output_height) {
			(void) jpeg_read_scanlines(&cinfo, &buf8, 1);
		  p8=buf8;
			for (i=0; i<iMax;i++) {	
			bufImage[0]=p8[c2];
			bufImage[c1]=p8[c1];
			bufImage[c2]=p8[0];
			bufImage+=spp; p8+=spp;
		  }
		 }
	  }
	  free(buf8);  
	  retval=cinfo.output_scanline;
	  (void) jpeg_finish_decompress(&cinfo);
  }
  catch (int err)
  {
	  retval = -err;
  }
  jpeg_destroy_decompress(&cinfo);
  if (infile) fclose(infile);
  return retval;
}



static JOCTET *thumb_compdata; static int thumb_datalen=0; 

static void thumbnail_src_init(struct jpeg_decompress_struct *cinfo) {
   // struct jpeg_state *h  = container_of(cinfo, struct jpeg_state, cinfo);
    cinfo->src->next_input_byte =thumb_compdata;
    cinfo->src->bytes_in_buffer = thumb_datalen;
//	cinfo->src->bytes_in_buffer=0; 
//	fprintf(stderr, "thumbnail_src_init\n");
}

static boolean thumbnail_src_fill(struct jpeg_decompress_struct *cinfo) {
/*	fprintf(stderr, "thumbnail_src_fill\n");
    cinfo->src->next_input_byte =thumb_compdata;
    cinfo->src->bytes_in_buffer = thumb_datalen;
	return 1; */
	exit(1);
}

static void thumbnail_src_skip(struct jpeg_decompress_struct *cinfo,long num_bytes) {
    cinfo->src->next_input_byte += num_bytes;
}

static void thumbnail_src_term(struct jpeg_decompress_struct *cinfo) {
}

static struct jpeg_source_mgr thumbnail_mgr;

static jmp_buf jpeglib_erreur;

static void gere_jpeglib_erreur(j_common_ptr cinfo) {
       longjmp(jpeglib_erreur, 1);
}

int __stdcall pmread_jpg_thumbnail (unsigned char *thumbCompData, int datalen, int queryType, int *thumb_width, int *thumb_height, int *thumb_spp, int *thumb_row, unsigned char *thumbDecompData) {
//queryType 1: image dim returned in pointers thumb_width, thumb_height and thumb_spp
//queryType 0: image data returned in thumbDecompData
//queryType 2: image data returned in thumbDecompData and height after black lines removed (if any) returned in thumb_height
	
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	unsigned char *buf8;
	int ligne=0, rowStride=0, icol=0, spp=1, c0=0, c1=0, c2=0, isBGR=0;
    int hasBlackLines=1, topBlackLines=0, bottomBlackLines=0; const int DarkVal=3*10;

     if (setjmp(jpeglib_erreur)) {  fprintf(stderr,"jpeglib call from pmread_jpg_thumbnail returned an error\n");	 return -3;	}

	cinfo.err = jpeg_std_error (&jerr);
	jerr.error_exit=gere_jpeglib_erreur;
	jpeg_create_decompress (&cinfo);
	
	thumb_compdata=thumbCompData;  thumb_datalen=datalen;

	thumbnail_mgr.init_source         = thumbnail_src_init,
    thumbnail_mgr.fill_input_buffer   = thumbnail_src_fill,
    thumbnail_mgr.skip_input_data     = thumbnail_src_skip,
    thumbnail_mgr.resync_to_restart   = jpeg_resync_to_restart,
    thumbnail_mgr.term_source         = thumbnail_src_term,

	cinfo.src = &thumbnail_mgr;

	jpeg_read_header(&cinfo, TRUE);
	jpeg_calc_output_dimensions(&cinfo); //fprintf(stderr, "thumb dim:  %d x %d\n", cinfo.output_width,cinfo.output_height);

    spp=cinfo.output_components;  c1=spp/2;  c2=spp-1;
    
	if (queryType==1) { //in this case, call to pmread_jpg_thumbnail is only to get dimensions of thumbnail
		*thumb_width=cinfo.output_width;  *thumb_height=cinfo.output_height;  *thumb_spp=spp; 
		jpeg_destroy_decompress(&cinfo);
		return cinfo.output_height;
	}	
    
    
	jpeg_start_decompress(&cinfo);

	rowStride=cinfo.output_width*spp;
    
    if (thumb_row!=NULL) {
        if (thumb_row[0]>0) {
            isBGR=1;
            c0=c2;  c2=0;
        }
        if (thumb_row[0]>rowStride) rowStride=thumb_row[0];
    }
    
	buf8=(unsigned char*)calloc(rowStride,sizeof(unsigned char));  if (buf8==NULL) return -4;

    hasBlackLines=(queryType==2);
    
    if (_debugThumb) fprintf(stderr, "pmread_jpg_thumbnail:  %d x %d  spp: %d  -  queryType: %d\n",cinfo.output_width, cinfo.output_height, spp, queryType);

	while (cinfo.output_scanline < cinfo.output_height) {

		(void) jpeg_read_scanlines(&cinfo, &buf8, 1);
        
        if (isBGR) {
            icol=0;
            while (icol<cinfo.output_width*spp) {
                thumbDecompData[ligne+icol+c0]=buf8[icol+c2];
                thumbDecompData[ligne+icol+c1]=buf8[icol+c1];
                thumbDecompData[ligne+icol+c2]=buf8[icol+c0];
                icol+=spp;
            }
        }
        else
            memcpy (thumbDecompData+ligne, buf8, rowStride);

       if (_debugThumb) if (cinfo.output_scanline<10 || (cinfo.output_scanline==(cinfo.output_height/2)) || (cinfo.output_scanline==cinfo.output_height) ) fprintf(stderr,
       "pmread_jpg_thumbnail - row %d  -  output:  %d %d %d -  %d %d %d -  %d %d %d\n",cinfo.output_scanline,
       buf8[0],buf8[1],buf8[2],buf8[rowStride/2],buf8[rowStride/2+1],buf8[rowStride/2+2],buf8[rowStride-spp],buf8[rowStride-spp+1],buf8[rowStride-spp+2]);


        if (hasBlackLines) {
            
                icol=0;
                while (icol<rowStride) {
                
                    if ((buf8[icol]+buf8[icol+c1]+buf8[icol+c2]) > DarkVal) {
                        if (_debugThumb) fprintf(stderr, "pmread_jpg_thumbnail:  no black at row: %d   col: %d  (blackLines: %d)  - val: %d %d %d\n",cinfo.output_scanline-1,icol,topBlackLines,buf8[icol],buf8[icol+c1],buf8[icol+c2]);
                        hasBlackLines=0;
                        break;
                    }
                    icol+=spp;
                }

                if (hasBlackLines) topBlackLines++;
            
        }

        if ( (topBlackLines>0) && (cinfo.output_scanline>(cinfo.output_height-(topBlackLines+2))) ) {
            icol=0;
            while (icol<rowStride) {
                if ((buf8[icol]+buf8[icol+c1]+buf8[icol+c2]) > DarkVal) break;
                icol+=spp;
            }

            if (_debugThumb) fprintf(stderr, "  row: %d   black pix till col: %d  (%d)\n",cinfo.output_scanline-1,icol,icol/spp);

            if (icol==rowStride) bottomBlackLines++;
            else {
                if (bottomBlackLines>0) bottomBlackLines=0;
            }
        }
        
		ligne+=rowStride;
   }

   free(buf8);	jpeg_finish_decompress(&cinfo); 	jpeg_destroy_decompress(&cinfo);
    
   if (queryType==2) {
        if ( ((topBlackLines>0)||(bottomBlackLines>0)) &&  ( (topBlackLines+bottomBlackLines) < (cinfo.output_height/4) ) ) {
            
            *thumb_height=cinfo.output_height-(topBlackLines+bottomBlackLines);
            
            memmove(thumbDecompData, thumbDecompData + rowStride * topBlackLines, rowStride*thumb_height[0]);
        }
	}
    else *thumb_height=cinfo.output_height;

    return ligne;
}



